import os
import shlex

from pkg_resources import resource_filename

from .config import CONFIG, TOOL_MAKE
from .terminal import Style
from .utils import execute, ExecOutput


class Extension:
    def __init__(self, work_dir):
        self.work_dir = work_dir

    def make(self, pg_config, targets=None, options=None):

        if not targets:
            targets = CONFIG['pgxs']['default_targets']

        if not options:
            options = CONFIG['pgxs']['default_options']

        # copy options
        opts = [x for x in options]

        # append compiler options, if needed (e.g. for scan_build)
        for env in ['CC', 'CXX']:
            if env in os.environ:
                opts.append('{}={}'.format(env, os.environ.get(env)))

        for target in targets:
            print(Style.green('$ {} {} {}').format(TOOL_MAKE,
                                                   target,
                                                   ' '.join(opts)))

            args = [
                TOOL_MAKE,
                'USE_PGXS=1',
                'PG_CONFIG={}'.format(pg_config),
                target,
            ] + opts

            # execute make (writes to stdout)
            execute(args,
                    cwd=self.work_dir,
                    env=os.environ,
                    output=ExecOutput.Stdout)
            print()

    def makefile_print_var(self, pg_config, name):
        makefile = os.path.join(self.work_dir, 'Makefile')
        print_mk = resource_filename('norsu', 'data/print.mk')

        args = [
            TOOL_MAKE,
            'USE_PGXS=1',
            'PG_CONFIG={}'.format(pg_config),
            '-f', makefile,
            '-f', print_mk,
            'print-{}'.format(name)
        ]

        key, _, value = execute(args).partition('=')
        return value

    def extra_regress_opts(self, pg_config):
        name = 'EXTRA_REGRESS_OPTS'
        ret = self.makefile_print_var(pg_config, name)

        result = {}
        for s in shlex.split(ret):
            key, _, value = s.partition('=')
            result[key] = value

        return result



import os
import subprocess

from enum import Enum

from .exceptions import Error
from itertools import tee, filterfalse


class ExecOutput(Enum):
    Stdout = None
    Pipe = subprocess.PIPE
    Devnull = subprocess.DEVNULL


def execute(args, cwd=None, env=None, error=True, output=ExecOutput.Pipe):
    p = subprocess.Popen(args,
                         cwd=cwd,
                         env=env,
                         stdout=output.value,
                         stderr=subprocess.STDOUT)

    if output == ExecOutput.Pipe:
        out, _ = p.communicate()
        out = out.decode('utf8')
    else:
        p.wait()
        out = None

    if p.returncode != 0:
        if error:
            raise Error('Failed to execute {}'.format(' '.join(args)))
    else:
        return out


def partition(pred, iterable):
    t1, t2 = tee(iterable)
    return filterfalse(pred, t1), filter(pred, t2)


def try_read_file(path):
    if os.path.exists(path):
        with open(path, 'r') as f:
            return f.read()

    return ''



import os
import sys

from shutil import rmtree
from testgres import get_new_node

from .config import NORSU_DIR, WORK_DIR, CONFIG
from .exceptions import Error
from .extension import Extension
from .git import find_relevant_refs
from .instance import Instance, InstanceName, sort_refs
from .terminal import Style
from .utils import partition, try_read_file


def extract_instances(args, dir):
    entries, options = extract_entries(args)

    entries_pos, entries_neg = partition(lambda x: x.startswith('^'), entries)

    entries_pos = list(entries_pos)
    entries_neg = [e[1:] for e in entries_neg]  # remove caps

    if not entries_pos:
        entries_pos = sorted([
            e for e in os.listdir(dir)
            if not e.startswith('.')
        ])

    entries = [e for e in entries_pos if e not in entries_neg]

    return (entries, options)


def extract_entries(args):
    entries, options = partition(lambda x: x.startswith('-'), args)

    entries = list(entries)
    options = list(options)

    return (entries, options)


def split_args(args):
    """
    Separate main args from auxiliary ones.
    """

    try:
        i = args.index('--')
        return (args[:i], args[i + 1:])
    except ValueError:
        return (args, [])


def cmd_instance(cmd, args):
    entries, _ = extract_instances(split_args(args)[0], NORSU_DIR)

    # safety pin (see config)
    if not args and cmd == 'remove' and \
       CONFIG['commands']['remove']['require_args']:
        raise Error('By default, this command requires arguments')

    for entry in entries:
        print('Selected instance:', Style.bold(entry))

        instance = Instance(entry)

        cmds = {
            'install': lambda: instance.install(),
            'remove': lambda: instance.remove(),
            'status': lambda: instance.status(),
            'pull': lambda: instance.pull(),
        }

        # execute command
        cmds[cmd]()

        print()


def cmd_search(_, args):
    entries, _ = extract_instances(split_args(args)[0], NORSU_DIR)

    for entry in entries:
        name = InstanceName(entry)
        patterns = name.to_patterns()

        print('Search query:', Style.bold(entry))

        refs = find_relevant_refs(CONFIG['repos']['urls'], patterns)

        for ref in sort_refs(refs, name):
            print('\t', ref.name)

        print()


def cmd_purge(_, args):
    entries, _ = extract_instances(split_args(args)[0], WORK_DIR)

    for entry in entries:
        instance = Instance(entry)

        if not os.path.exists(instance.main_dir):
            rmtree(path=instance.work_dir, ignore_errors=True)


def cmd_pgxs(_, args):
    main_args, make_args = split_args(args)

    pgs, cmd_opts = extract_instances(main_args, NORSU_DIR)
    targets, make_opts = extract_entries(make_args)
    work_dir = os.getcwd()

    # extension we're going to build
    extension = Extension(work_dir=work_dir)

    for pg in pgs:
        instance = Instance(pg)
        pg_config = instance.get_bin_path('pg_config')

        if instance.installed_commit_hash:
            print('Executing against instance', Style.bold(pg), '\n')
        else:
            print(Style.yellow('Cannot find instance {}\n'.format(pg)))
            continue

        # should we start PostgreSQL?
        if any(k in cmd_opts for k in ['-R', '--run-pg']):
            regress_opts = extension.extra_regress_opts(pg_config)
            temp_conf_file = regress_opts.get('--temp-config')
            temp_conf = ''

            # read additional config
            if temp_conf_file:
                path = os.path.join(work_dir, temp_conf_file)
                temp_conf = try_read_file(path)
                print('Found custom config:', os.path.basename(path))

            # run commands under a running PostgreSQL instance
            with get_new_node(port=5432) as node:
                print('Starting temporary PostgreSQL instance...\n')
                os.environ['PG_CONFIG'] = pg_config
                node.cleanup_on_bad_exit = True
                node.init().append_conf(line=temp_conf).start()

                extension.make(pg_config, targets, make_opts)
        else:
            extension.make(pg_config, targets, make_opts)

        print()  # splitter


def cmd_path(_, args):
    entries, _ = extract_instances(split_args(args)[0], NORSU_DIR)

    for entry in entries:
        print(Instance(entry).main_dir)


def cmd_help(*_):
    name = os.path.basename(sys.argv[0])
    print('{} -- PostgreSQL builds manager'.format(Style.blue(name)))
    print()
    print('Usage:')
    print('\t{} <command> [options]'.format(name))
    print()
    print('Commands:')

    for method in METHODS.keys():
        print('\t{}'.format(method))

    print()
    print('Examples:')
    print('\t{} install 9.6.5 10 master'.format(name))
    print('\t{} pgxs 9.6 9.5 -- install'.format(name))
    print('\t{} pull REL_10_STABLE'.format(name))
    print('\t{} remove 9.5'.format(name))
    print('\t{} status'.format(name))


def main():
    args = sys.argv[1:]
    if len(args) == 0:
        args = ['install']

    command = args[0]
    method = METHODS.get(command)

    try:
        if method is None:
            raise Error('Unknown command {}'.format(command))
        method(command, args[1:])
    except KeyboardInterrupt:
        pass
    except Error as e:
        print(Style.red(str(e)))
        exit(1)


METHODS = {
    'install': cmd_instance,
    'remove': cmd_instance,
    'status': cmd_instance,
    'search': cmd_search,
    'pull': cmd_instance,
    'purge': cmd_purge,
    'pgxs': cmd_pgxs,
    'path': cmd_path,
    'help': cmd_help,
}



import os
import re
import shlex

from enum import Enum
from shutil import rmtree

from .config import NORSU_DIR, WORK_DIR, CONFIG, TOOL_MAKE
from .exceptions import Error
from .terminal import Style
from .utils import execute, ExecOutput

from .git import \
    GitRepo, \
    SortRefByVersion, \
    SortRefBySimilarity, \
    find_relevant_refs


def step(*args):
    print(Style.green('\t=>'), *args)


def line(name, value=None):
    print('\t', name, '\t{}'.format(value) if value is not None else '')


def read_commit_file(path):
    if os.path.exists(path):
        with open(path, 'r') as f:
            return f.read().strip()


def write_commit_file(path, value):
    with open(path, 'w') as f:
        f.write(value)


def sort_refs(refs, name):
    # key function for sort
    def to_key(x):
        if name.type == InstanceNameType.Version:
            return SortRefByVersion(x)
        else:
            # pre-calculated for better performance
            name_ngram = SortRefBySimilarity.ngram(name.value)
            return SortRefBySimilarity(x, name_ngram)

    return sorted(refs, reverse=True, key=to_key)


class InstanceNameType(Enum):
    Version = 1
    Branch = 2


class InstanceName:
    rx_is_ver = re.compile(r'\d+([._]\d+)*')
    rx_sep = re.compile(r'(\.|_)')

    def __init__(self, name):
        for s in ['/']:
            if s in name:
                raise Error('Wrong name {}'.format(name))

        self.value = name

        if self.rx_is_ver.match(name):
            self.type = InstanceNameType.Version
        else:
            self.type = InstanceNameType.Branch

    def to_patterns(self):
        pattern = self.value
        result = [pattern]

        if self.type == InstanceNameType.Version:
            # replace version separators with a pattern
            pattern = self.rx_sep.sub(lambda m: '[._]', pattern)

            for fmt in ['REL_{}*', 'REL{}*']:
                result.append(fmt.format(pattern))
        else:
            result.append('*{}*'.format(pattern))

        return result

    def __str__(self):
        return self.value


class Instance:
    def __init__(self, name):
        self.name = InstanceName(name)
        self.main_dir = os.path.join(NORSU_DIR, name)
        self.work_dir = os.path.join(WORK_DIR, name)
        self.git = GitRepo(work_dir=self.work_dir)

        # various utility files
        self.configure_file = os.path.join(self.main_dir, '.norsu_configure')
        self.ignore_file = os.path.join(self.main_dir, '.norsu_ignore')

        # store commit hashes (build + install)
        self.installed_commit_file = os.path.join(self.main_dir, '.norsu_build')
        self.built_commit_file = os.path.join(self.work_dir, '.norsu_build')

    @property
    def ignore(self):
        return os.path.exists(self.ignore_file)

    @property
    def actual_commit_hash(self):
        if os.path.exists(self.work_dir):
            return self.git.hash

    @property
    def installed_commit_hash(self):
        return read_commit_file(self.installed_commit_file)

    @installed_commit_hash.setter
    def installed_commit_hash(self, value):
        write_commit_file(self.installed_commit_file, value)

    @property
    def built_commit_hash(self):
        return read_commit_file(self.built_commit_file)

    @built_commit_hash.setter
    def built_commit_hash(self, value):
        write_commit_file(self.built_commit_file, value)

    @property
    def requires_reinstall(self):
        # NOTE: remember that re-build != re-install!
        # We might already have a fresh build in work dir
        return self.installed_commit_hash != self.actual_commit_hash

    @property
    def requires_rebuild(self):
        return self.built_commit_hash != self.actual_commit_hash

    def get_bin_path(self, name):
        return os.path.join(self.main_dir, 'bin', name)

    def pg_config(self, params=None):
        pg_config = self.get_bin_path('pg_config')
        if os.path.exists(pg_config):
            return execute([pg_config] + params)

    def status(self):
        postgres = os.path.join(self.main_dir, 'bin', 'postgres')

        if os.path.exists(postgres):
            if self.requires_reinstall:
                status = Style.yellow('Installed (out of date)')
            else:
                status = Style.green('Installed')
        else:
            status = Style.red('Not installed')

        line('Status:', status)

        if os.path.exists(self.main_dir):
            line('Main dir:', self.main_dir)
        else:
            line('Main dir:', 'none')

        if os.path.exists(self.work_dir):
            line('Work dir:', self.work_dir)
            branch = self.git.branch or self.git.tag
            if branch:
                line('Branch:', branch)
        else:
            line('Work dir:', 'none')

        pg_config_out = self.pg_config(['--version'])
        if pg_config_out:
            line('Version:', pg_config_out.strip())

        commit = self.installed_commit_hash
        if commit:
            line('Commit:', commit)

        configure = self._configure_options()
        line('CONFIGURE:', configure)

    def pull(self):
        self._prepare_work_dir()

    def install(self):
        if not self.ignore:
            try:
                self._prepare_work_dir()
                self._make_distclean()
                self._configure_project()
                self._make_install()
            except Error as e:
                step(Style.red(str(e)))
        else:
            step(Style.yellow('Ignored due to .norsu_ignore'))

    def remove(self):
        for path, name in [(self.main_dir, 'main'), (self.work_dir, 'work')]:
            if os.path.exists(path):
                rmtree(path=path, ignore_errors=True)
                step('Removed {} dir'.format(name))

    def _configure_options(self):
        if os.path.exists(self.configure_file):
            with open(self.configure_file, 'r') as f:
                return shlex.split(f.read())

        pg_config_out = self.pg_config(['--configure'])
        if pg_config_out:
            options = shlex.split(pg_config_out)
            return [x for x in options if not x.startswith('--prefix')]

        return CONFIG['build']['configure_options']

    def _prepare_work_dir(self):
        git_repo = os.path.join(self.work_dir, '.git')

        if os.path.exists(git_repo):
            branch = self.git.branch

            # pull latest changes
            if branch:
                self.git.pull()

            # should we reinstall PG?
            if self.requires_reinstall:
                fresh_commits = ''

                # show distance between installed and fresh commits
                installed_commit = self.installed_commit_hash
                if installed_commit:
                    commits = self.git.distance(installed_commit, branch)
                    fresh_commits = ' ({} commits)'.format(commits)

                step('Current branch:', Style.bold(branch))
                step('Installed build is out of date{}'.format(fresh_commits))
        else:
            step('No work dir, choosing repo & branch')

            patterns = self.name.to_patterns()
            refs = find_relevant_refs(CONFIG['repos']['urls'], patterns)

            if not refs:
                raise Error('No branch found for {}'.format(self.name))

            # select the most relevant branch
            ref = sort_refs(refs, self.name)[0]
            step('Selected repo', Style.bold(ref.repo))
            step('Selected branch', Style.bold(ref.name))

            # finally, clone repo
            self.git.clone(url=ref.repo, branch=ref.name)
            step('Cloned git repo to work dir')

        # add .norsu* to git excludes
        excludes = os.path.join(self.work_dir, '.git', 'info', 'exclude')
        with open(excludes, 'r+') as f:
            lines = f.readlines()
            if not any('.norsu*' in s for s in lines):
                f.seek(0, os.SEEK_END)
                f.write('.norsu*')

    def _configure_project(self):
        makefile = os.path.join(self.work_dir, 'GNUmakefile')
        if not os.path.exists(makefile):
            args = [
                './configure',
                '--prefix={}'.format(self.main_dir)
            ] + self._configure_options()

            execute(args, cwd=self.work_dir, output=ExecOutput.Devnull)
            step('Configured sources')

    def _make_install(self):
        if self.requires_reinstall:
            # update built commit hash
            self.built_commit_hash = self.actual_commit_hash

            jobs = int(CONFIG['build']['jobs'])
            for arg in ['-j{}'.format(jobs), 'install']:
                args = [TOOL_MAKE, arg]
                execute(args, cwd=self.work_dir, output=ExecOutput.Devnull)

            # update installed commit hash
            self.installed_commit_hash = self.actual_commit_hash

            step('Built and installed')

    def _make_distclean(self):
        makefile = os.path.join(self.work_dir, 'GNUmakefile')
        if os.path.exists(makefile) and self.requires_rebuild:
            args = [TOOL_MAKE, 'distclean']
            execute(args, cwd=self.work_dir, error=False,
                    output=ExecOutput.Devnull)

            step('Prepared work dir for a new build')
